#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/**** sample: interval timer :begin ****/
#include <time.h>
#include <sys/time.h>
/**** sample: interval timer :end ****/

#define MAIN_C
#include "exitfail.h"

#define SERIAL_BAUDRATE B115200
#define BUF_SIZE 2048

static int serial_fd = -1;
static struct termios old_tio;

static int terminated = 0;

#define BASENAME(p) ((strrchr((p), '/') ? : ((p) - 1)) +1)
#define ARRAY_SIZE(a) (sizeof(a) /sizeof((a)[0]))

#define __unused __attribute__((unused))

/**** sample: RN4020 command: begin ****/
static unsigned char cmd_sf[] = "SF,1\n";
static unsigned char cmd_ss[] = "SS,C0000000\n";
static unsigned char cmd_sr[] = "SR,92000000\n";
static unsigned char cmd_r[] = "R,1\n";
/**** sample: RN4020 command: begin ****/

/**** sample: RN4020 response: begin ****/
static unsigned char res_aok[] = "AOK";
static unsigned char res_reboot[] = "Reboot";
static unsigned char res_cmd[] = "CMD";
/**** sample: RN4020 response: end ****/

static void terminate_sig_handler(__unused int signum)
{
	terminated = 1;
}

static void set_sig_handler(int sig_list[], ssize_t num, __sighandler_t handler)
{
	struct sigaction sa;
	int i;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;

	for(i=0; i<num; i++)
		if(sigaction(sig_list[i], &sa, NULL) < 0)
			exitfail_errno("sigaction");
}

static void restore_serial(void)
{
	int ret;

	ret = tcsetattr(serial_fd, TCSANOW, &old_tio);
	if(ret < 0) exitfail_errno("tcsetattr");
}

static void setup_serial(int fd)
{
	struct termios tio;
	int ret;

	ret = tcgetattr(fd, &old_tio);
	if(ret) exitfail_errno("tcsgetattr");

	if(atexit(restore_serial)) exitfail_errno("atexit");

	memset(&tio, 0, sizeof(tio));
	tio.c_iflag = IGNBRK | IGNPAR;
	tio.c_cflag = CS8 | CLOCAL | CREAD;
	tio.c_cc[VTIME] = 0;
	tio.c_cc[VMIN] = 1;
	ret = cfsetspeed(&tio, SERIAL_BAUDRATE);
	if(ret < 0) exitfail_errno("cfsetspeed");

	ret = tcflush(fd, TCIFLUSH);
	if(ret < 0) exitfail_errno("tcflush");

	ret = tcsetattr(fd, TCSANOW, &tio);
	if(ret) exitfail_errno("tcsetattr");
}

int main(int argc, char *argv[])
{
	int terminate_sig_list[] = {
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};

/**** sample: variable: begin ****/
	ssize_t ret;
       unsigned char buf[BUF_SIZE];
       unsigned char buf1[BUF_SIZE];
	int i;
	int flag_aok, flag_reboot, flag_cmd;
	ssize_t analen;
	int cnt_cmp;
/**** sample: variable: end ****/

/**** sample: interval timer: begin ****/
	struct itimerval itimer;
/**** sample: interval timer: end ****/

	exitfail_init();

	if(argc < 2) {
		printf("Usage: %s <device>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	serial_fd = open(argv[1], O_RDWR | O_NOCTTY);
	if(serial_fd < 0) exitfail_errno("open");

	set_sig_handler(terminate_sig_list, ARRAY_SIZE(terminate_sig_list),
			terminate_sig_handler);

	setup_serial(serial_fd);

/**** sample: interval timer: begin ****/
	itimer.it_value.tv_sec = itimer.it_interval.tv_sec = 5;
	itimer.it_value.tv_usec = itimer.it_interval.tv_usec = 0;
/**** sample: interval timer: end ****/

/**** sample: RN4020 command(SF): begin ****/
	ret = write(serial_fd, cmd_sf, sizeof(cmd_sf) - 1);
	if (ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: %s",cmd_sf);

	flag_aok = 0;
	cnt_cmp = 0;
	analen = 0;

	while(!terminated && flag_aok != 1) {
		if(setitimer(ITIMER_REAL, &itimer, NULL) < 0) {
			exitfail_errno("setitimer");
			return 0;
		}
	
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");

		for(i = 0; i < ret && flag_aok != 1; i++) {
			printf("%c", buf[i]);
			/* Not considered about search resoponse. ex."AOAOK" */
			if(buf[i] == res_aok[cnt_cmp]) {
				cnt_cmp++;
				buf1[analen++] = buf[i];
			}
			else {
				cnt_cmp = 0;
				analen = 0;
			}
			if(cnt_cmp == sizeof(res_aok) - 1) {
				flag_aok = 1;
				buf1[analen] = '\0';
			}
		}
		if(flag_aok == 1) {
			printf("\n-------- Receive: %s\n", res_aok);
//			printf("%s\n", buf1);
		}
	}
/**** sample: RN4020 command(SF): end ****/

/**** sample: RN4020 command(SS): begin ****/
	ret = write(serial_fd, cmd_ss, sizeof(cmd_ss) - 1);
	if (ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: %s", cmd_ss);

	flag_aok = 0;
	cnt_cmp = 0;
	analen = 0;

	while(!terminated && flag_aok != 1) {
		if(setitimer(ITIMER_REAL, &itimer, NULL) < 0) {
			exitfail_errno("setitimer");
			return 0;
		}
	
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");

		for(i = 0; i < ret && flag_aok != 1; i++) {
			printf("%c", buf[i]);
			/* Not considered about search resoponse. ex."AOAOK" */
			if(buf[i] == res_aok[cnt_cmp]) {
				cnt_cmp++;
				buf1[analen++] = buf[i];
			}
			else {
				cnt_cmp = 0;
				analen = 0;
			}
			if(cnt_cmp == sizeof(res_aok) - 1) {
				flag_aok = 1;
				buf1[analen] = '\0';
			}
		}
		if(flag_aok == 1) {
			printf("\n-------- Receive: %s\n", res_aok);
//			printf("%s\n", buf1);
		}
	}
/**** sample: RN4020 command(SS): end ****/

/**** sample: RN4020 command(SR): begin ****/
	ret = write(serial_fd, cmd_sr, sizeof(cmd_sr) - 1);
	if (ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: %s", cmd_sr);

	flag_aok = 0;
	cnt_cmp = 0;
	analen = 0;

	while(!terminated && flag_aok != 1) {
		if(setitimer(ITIMER_REAL, &itimer, NULL) < 0) {
			exitfail_errno("setitimer");
			return 0;
		}
	
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");

		for(i = 0; i < ret && flag_aok != 1; i++) {
			printf("%c", buf[i]);
			/* Not considered about search resoponse. ex."AOAOK" */
			if(buf[i] == res_aok[cnt_cmp]) {
				cnt_cmp++;
				buf1[analen++] = buf[i];
			}
			else {
				cnt_cmp = 0;
				analen = 0;
			}
			if(cnt_cmp == sizeof(res_aok) - 1) {
				flag_aok = 1;
				buf1[analen] = '\0';
			}
		}
		if(flag_aok == 1) {
			printf("\n-------- Receive: %s\n", res_aok);
//			printf("%s\n", buf1);
		}
	}
/**** sample: RN4020 command(SR): end ****/

/**** sample: RN4020 command(R): begin ****/
	ret = write(serial_fd, cmd_r, sizeof(cmd_r) - 1);
	if (ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: %s", cmd_r);

	flag_reboot = 0;
	flag_cmd = 0;
	cnt_cmp = 0;
	analen = 0;

	while(!terminated && flag_cmd != 1) {
		if(setitimer(ITIMER_REAL, &itimer, NULL) < 0) {
			exitfail_errno("setitimer");
			return 0;
		}
	
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");

		for(i = 0; i < ret && flag_cmd != 1; i++) {
			printf("%c", buf[i]);
			if(flag_reboot != 1) {
				/* Not considered about search resoponse. ex."ReReboot" */
				if(buf[i] == res_reboot[cnt_cmp]) {
					cnt_cmp++;
					buf1[analen++] = buf[i];
				}
				else {
					cnt_cmp = 0;
					analen = 0;
				}
				if(cnt_cmp == sizeof(res_reboot) - 1) {
					flag_reboot = 1;
					buf1[analen] = '\0';
				}
			}
			else {
				/* Not considered about search resoponse. ex."CMCMD" */
				if(buf[i] == res_cmd[cnt_cmp]) {
					cnt_cmp++;
					buf1[analen++] = buf[i];
				}
				else {
					cnt_cmp = 0;
					analen = 0;
				}
				if(cnt_cmp == sizeof(res_cmd) - 1) {
					flag_cmd = 1;
					buf1[analen] = '\0';
				}
			}
		}
		if(flag_reboot == 1 &&flag_cmd == 1)
			printf("\n-------- Receive: %s and %s\n", res_reboot, res_cmd);
	}
/**** sample: RN4020 command(R): end ****/
	printf("\n######## Finish RN4020 Setting ########\n");

	return EXIT_SUCCESS;
}

